from .model import Model
from .._lsuv import do_lsuv
from ... import describe
from ...describe import Weights, Dimension, Gradient
from ..util import copy_array
import random
import numpy

def LSUVinit(model, X, y=None):
    if model.vectors is not None:
        do_lsuv(model.ops, model.vectors, model, X)
    return X

def _uniform_init(lo, hi):
    def wrapped(W, ops):
        if (W**2).sum() == 0.:
            copy_array(W, ops.xp.random.uniform(lo, hi, W.shape))
    return wrapped


#@describe.on_data(LSUVinit)
@describe.attributes(
    nO=Dimension("Vector dimensions"),
    nV=Dimension("Number of vectors"),
    vectors=Weights("Embedding table",
        lambda obj: (obj.nV, obj.nO),
        _uniform_init(-0.1, 0.1)
    ),
    d_vectors=Gradient("vectors"),
)
class HashEmbed(Model):
    name = 'hash-embed'
    def __init__(self, nO, nV, seed=None, **kwargs):
        Model.__init__(self, **kwargs)
        self.column = kwargs.get('column', 0)
        self.nO = nO
        self.nV = nV
        self.seed = self.id

    def predict(self, ids):
        if ids.ndim >= 2:
            ids = self.ops.xp.ascontiguousarray(ids[:, self.column], dtype='uint64')
        keys = self.ops.hash(ids, self.seed) % self.nV
        vectors = self.vectors[keys]
        summed = vectors.sum(axis=1)
        return summed

    def begin_update(self, ids, drop=0.):
        if ids.ndim >= 2:
            ids = self.ops.xp.ascontiguousarray(ids[:, self.column], dtype='uint64')
        keys = self.ops.hash(ids, self.seed) % self.nV
        vectors = self.vectors[keys].sum(axis=1)
        mask = self.ops.get_dropout_mask((vectors.shape[1],), drop)
        if mask is not None:
            vectors *= mask
        def finish_update(delta, sgd=None):
            if mask is not None:
                delta *= mask
            keys = self.ops.hash(ids, self.seed) % self.nV
            d_vectors = self.d_vectors
            keys = self.ops.xp.ascontiguousarray(keys.T, dtype='i')
            for i in range(keys.shape[0]):
                self.ops.scatter_add(d_vectors, keys[i], delta)
            if sgd is not None:
                sgd(self._mem.weights, self._mem.gradient, key=self.id)
            return None
        return vectors, finish_update



import pytest

from ...neural.mem import Memory
from ...neural.ops import NumpyOps

@pytest.fixture
def ops():
    return NumpyOps()

@pytest.mark.parametrize('size', [0, 10, 1000, 7, 12])
def test_init_allocates_mem(ops, size):
    params = Memory(ops, size)
    assert params._mem[0].size == size
    assert params._i == 0

@pytest.mark.parametrize('size', [-10, -1000, -1, -7, -12])
def test_init_rejects_negative_sizes(ops, size):
    with pytest.raises(ValueError):
        params = Memory(ops, size)

def test_add_param_within_size(ops):
    params = Memory(ops, size=128)
    params.add('W', (5, 10))
    assert params._offsets['W'] == (0, 0, (5, 10))
    params.add('b', (5,))
    assert params._offsets['b'] == (5*10, 0, (5,))
    

def test_add_param_realloc(ops):
    params = Memory(ops, size=10)
    params.add('b', (5,))
    assert params._offsets['b'] == (0, 0, (5,))
    params.add('W', (5, 10))
    assert params._offsets['W'] == (5, 0, (5, 10))
    assert params._offsets['b'] == (0, 0, (5,))
 

def test_get_param_present(ops):
    params = Memory(ops, size=10)
    b = params.add('b', (5,))
    b2 = params.get('b')
    b[0] = 100
    assert b[0] == b2[0]
 

def test_get_param_absent(ops):
    params = Memory(ops, size=10)
    b = params.get('b')
    assert b is None
 

@pytest.mark.xfail
def test_get_first_gradient(ops):
    params = Memory(ops, size=10)
    b = params.add('b', (5,))
    b2 = params.get('d_b')
    b[0] = 100
    assert b2[0] == 0
 

@pytest.mark.xfail
def test_get_existing_gradient(ops):
    params = Memory(ops, size=10)
    b = params.add('b', (5,))
    b2 = params.get('d_b')
    b[0] = 100
    assert b2[0] == 0
    b2[0] = 20.
    b3 = params.get('d_b')
    assert b3[0] == b2[0]
 

def test_get_gradient_absent_parameter(ops):
    params = Memory(ops, size=10)
    d_b = params.get('d_b')
    assert d_b is None


#def test_merge_empty_others(ops):
#    params = Memory(ops, size=10)
#    assert params.allow_resize
#    params.merge_params([])
#    assert params.allow_resize
#
#
#def test_merge_no_resize(ops):
#    parent = Memory(ops, size=5)
#    assert parent.allow_resize
#    child = Memory(ops, size=2)
#    w_parent = parent.add('W', (4,))
#    w_child = child.add('W', (2,))
#    child._mem[0, 0] = 10.0
#    assert parent._i == 4
#    assert child._i == 2
#    parent.merge_params([child])
#    assert not parent.allow_resize
#    assert not child.allow_resize
#    assert parent._i == 6
#    assert child._i == 2
#
#
#def test_merge_with_resize(ops):
#    parent = Memory(ops, size=5)
#    child = Memory(ops, size=5)
#    w_parent = parent.add('W', (4,))
#    w_parent[0] += 2.
#    w_child = child.add('W', (3,))
#    w_child[0] += 5.
#    parent.merge_params([child])
#    w_parent = parent.get('W')
#    w_child = child.get('W')
#    assert w_child[0] == 5
#    assert w_parent[0] == 2
#
#
#def test_resize_disallowed_after_merge(ops):
#    parent = Memory(ops, size=5)
#    child = Memory(ops, size=5)
#    w_parent = parent.add('W', (4,))
#    w_child = child.add('W', (3,))
#    parent.merge_params([child])
#    with pytest.raises(ValueError):
#        child.add('b', (2,))
#
#def test_merge_disallowed_after_merge(ops):
#    parent = Memory(ops, size=5)
#    child = Memory(ops, size=5)
#    parent.add('W', (4,))
#    child.add('W', (3,))
#    parent.merge_params([child])
#    parent2 = Memory(ops, size=5)
#    with pytest.raises(ValueError):
#        parent2.merge_params([parent])
#    with pytest.raises(ValueError):
#        parent.replace_mem([parent2])
